@@ -602,9 +602,10 @@ def test_queue_mode_set_2_fail(re_manager, plist, exit_code):  # noqa F811
     (2, 2, True),
     (3, 2, True),
     (100, 2, True),
-    (-1, 1, True),
-    (-2, 0, True),
+    (-1, 2, True),
+    (-2, 1, True),
     (-3, 0, True),
+    (-4, 0, True),
     (-100, 0, True),
 ])
 # fmt: on
@@ -1217,17 +1218,20 @@ def test_function_execute_2_fail(re_manager):  # noqa: F811
     assert subprocess.call(["qserver", "function", "execute", item, "invalid_param"]) == PARAM_ERROR
 
 
-def test_task_result_1(re_manager):  # noqa: F811
+def test_task_result_status_1(re_manager):  # noqa: F811
     """
-    Tests for 'qserver task result_get'.
+    Tests for 'qserver task result' and 'qserver task status.
     """
     # The request should be successful for any 'task_uid'.
     task_uid = "01e80342-5e36-44de-bc86-9bd8d57c9885"
+    assert subprocess.call(["qserver", "task", "status", task_uid]) == SUCCESS
     assert subprocess.call(["qserver", "task", "result", task_uid]) == SUCCESS
 
     # Some cases of invalid parameters
+    assert subprocess.call(["qserver", "task", "status"]) == PARAM_ERROR
     assert subprocess.call(["qserver", "task", "result"]) == PARAM_ERROR
     assert subprocess.call(["qserver", "task", "something"]) == PARAM_ERROR
+    assert subprocess.call(["qserver", "task", "status", task_uid, "extra_param"]) == PARAM_ERROR
     assert subprocess.call(["qserver", "task", "result", task_uid, "extra_param"]) == PARAM_ERROR
 
 
@@ -1403,5 +1407,6 @@ def test_qserver_zmq_keys():
 
     # Generated public key based on private key - success
     _, private_key = generate_new_zmq_key_pair()
+    ttime.sleep(1)  # The pause may fix issues with this test
     print(f"Private key used for the test: '{private_key}'")
     assert subprocess.call(["qserver-zmq-keys", "--zmq-private-key", private_key]) == SUCCESS
